#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "resource.h"

#pragma pack(push,gifpacking,1)

typedef struct {
	char cSignature[3]; // Must be 'GIF'
	char cVersion[3];   // Must be '89a'
} GIF_HEADER;

typedef struct { // 7 bytes
	unsigned short iWidth;
	unsigned short iHeight;
	unsigned char iSizeOfGct : 3;
	unsigned char iSortFlag : 1;
	unsigned char iColorResolution : 3;
	unsigned char iGctFlag : 1;
	unsigned char iBackgroundColorIndex;
	unsigned char iPixelAspectRatio;
} GIF_LOGICAL_SCREEN_DESCRIPTOR;

typedef struct { // 6 bytes
	unsigned char iBlockSize;           // Must be 4
	unsigned char iTransparentColorFlag : 1;
	unsigned char iUserInputFlag : 1;
	unsigned char iDisposalMethod : 3;
	unsigned char iReserved : 3;
	unsigned short iDelayTime;
	unsigned char iTransparentColorIndex;
	unsigned char iBlockTerminator;     // Must be 0
} GIF_GRAPHIC_CONTROL_EXTENSION;

typedef struct { // 9 bytes
	unsigned short iLeft;
	unsigned short iTop;
	unsigned short iWidth;
	unsigned short iHeight;
	unsigned char iSizeOfLct : 3;
	unsigned char iReserved : 2;
	unsigned char iSortFlag : 1;
	unsigned char iInterlaceFlag : 1;
	unsigned char iLctFlag : 1;
} GIF_IMAGE_DESCRIPTOR;

#pragma pack(pop,gifpacking)

unsigned short iGctSize[]={6,12,24,48,96,192,384,768};

LRESULT WINAPI MainDlgProc(HWND,UINT,WPARAM,LPARAM);
void DropFiles(HDROP);
void UpdateFrameInfo(void);
void MakeGIF(void);

HINSTANCE hInst;
HANDLE hHeap;
HICON hIconMain;
HWND hMainDlg;

int WINAPI WinMain(HINSTANCE hInstance,
									 HINSTANCE hPrevInstance,
									 LPSTR lpszCmdLine,
									 int nShowCmd)
{
	INITCOMMONCONTROLSEX icce;
	icce.dwSize=sizeof(icce);
	icce.dwICC=ICC_UPDOWN_CLASS;
	InitCommonControlsEx(&icce);
	hInst=hInstance;
	hHeap=GetProcessHeap();
	hIconMain=LoadIcon(hInstance,MAKEINTRESOURCE(IDI_MAIN));
	DialogBox(hInstance,MAKEINTRESOURCE(IDD_MAIN),NULL,(DLGPROC)MainDlgProc);
	return FALSE;
}

LRESULT CALLBACK MainDlgProc(HWND hWnd,
														 UINT uMsg,
														 WPARAM wParam,
														 LPARAM lParam)
{
	DWORD dw;

	switch (uMsg) {
		case WM_INITDIALOG:
			hMainDlg=hWnd;
			SendMessage(hWnd,WM_SETICON,ICON_BIG,(LPARAM)hIconMain);
			CheckDlgButton(hWnd,IDC_CHECK_MAIN_LOOP,1);
			SendDlgItemMessage(hWnd,IDC_SPIN_MAIN_DELAY,UDM_SETRANGE,0,MAKELONG(1000,0));
			SendDlgItemMessage(hWnd,IDC_BUTTON_MAIN_WEBSITE,BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconMain);
			DragAcceptFiles(hWnd,TRUE);
			return TRUE;
		case WM_CLOSE:
			DragAcceptFiles(hWnd,FALSE);
			SendMessage(hWnd,WM_COMMAND,MAKELONG(IDC_BUTTON_MAIN_CLEARFRAMES,BN_CLICKED),0);
			EndDialog(hWnd,IDCANCEL);
			PostQuitMessage(0);
			return TRUE;
		case WM_COMMAND:
			switch (HIWORD(wParam)) {
				case BN_CLICKED:
					switch (LOWORD(wParam)) {
						case IDC_BUTTON_MAIN_REMOVEFRAME:
							dw=SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_GETCURSEL,0,0);
							if (dw!=-1) {
								HeapFree(hHeap,0,(LPVOID)SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_GETITEMDATA,dw,0));
								SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_DELETESTRING,dw,0);
							}
							if (!SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_GETCOUNT,0,0)) {
								EnableWindow(GetDlgItem(hWnd,IDC_BUTTON_MAIN_MAKEGIF),FALSE);
							}
							return TRUE;
						case IDC_BUTTON_MAIN_CLEARFRAMES:
							while (SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_GETCOUNT,0,0)) {
								HeapFree(hHeap,0,(LPVOID)SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_GETITEMDATA,0,0));
								SendDlgItemMessage(hWnd,IDC_LIST_MAIN_FRAMES,LB_DELETESTRING,0,0);
							}
							SetDlgItemText(hWnd,IDC_STATIC_MAIN_FILENAME,"");
							SetDlgItemText(hWnd,IDC_STATIC_MAIN_INFO,"");
							EnableWindow(GetDlgItem(hWnd,IDC_BUTTON_MAIN_MAKEGIF),FALSE);
							return TRUE;
						case IDC_BUTTON_MAIN_MAKEGIF:
							MakeGIF();
							return TRUE;
						case IDC_BUTTON_MAIN_WEBSITE:
							if ((long)ShellExecute(hWnd,NULL,"http://www.whitsoftdev.com/",NULL,NULL,SW_NORMAL)<=32)
								MessageBox(hWnd,"Error opening web browser. You may manually surf to http://www.whitsoftdev.com/.",NULL,MB_ICONEXCLAMATION);
							return TRUE;
						default:
							return FALSE;
					}
				case LBN_SELCHANGE:
					switch (LOWORD(wParam)) {
						case IDC_LIST_MAIN_FRAMES:
							UpdateFrameInfo();
							return TRUE;
						default:
							return FALSE;
					}
				default:
					return FALSE;
			}
		case WM_DROPFILES:
			DropFiles((HDROP)wParam);
			return TRUE;
		default:
			return FALSE;
	}
}

void DropFiles(HDROP hDrop)
{
	DWORD dw, dwTotal, dwLength;
	char *psz;

	dwTotal=DragQueryFile(hDrop,-1,0,0);
	for (dw=0;dw<dwTotal;dw++) {
		dwLength=DragQueryFile(hDrop,dw,0,0);
		psz=(char *)HeapAlloc(hHeap,0,dwLength+1);
		DragQueryFile(hDrop,dw,psz,dwLength+1);
		SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_SETITEMDATA,SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_ADDSTRING,0,(LPARAM)strrchr(psz,'\\')+1),(LPARAM)psz);
	}
	DragFinish(hDrop);
	EnableWindow(GetDlgItem(hMainDlg,IDC_BUTTON_MAIN_MAKEGIF),TRUE);
}

void UpdateFrameInfo()
{
	DWORD dwIndex, dw;
	char sz[256], c;
	char *pszFileName;
	HANDLE hFile;
	GIF_HEADER gh;
	GIF_LOGICAL_SCREEN_DESCRIPTOR glsd;
	GIF_IMAGE_DESCRIPTOR gid;
	ZeroMemory(&gh,sizeof(GIF_HEADER));
	ZeroMemory(&glsd,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR));
	ZeroMemory(&gid,sizeof(GIF_IMAGE_DESCRIPTOR));
	dwIndex=SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_GETCURSEL,0,0);
	SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_GETTEXT,dwIndex,(LPARAM)sz);
	pszFileName=(char *)SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_GETITEMDATA,dwIndex,0);
	SetDlgItemText(hMainDlg,IDC_STATIC_MAIN_FILENAME,sz);
	hFile=CreateFile(pszFileName,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0);
	if (hFile==INVALID_HANDLE_VALUE) {
		MessageBox(hMainDlg,"Could not open file.",0,MB_ICONEXCLAMATION);
		return;
	}
	ReadFile(hFile,&gh,sizeof(GIF_HEADER),&dw,0);
	if (strncmp(gh.cSignature,"GIF",3) || (strncmp(gh.cVersion,"89a",3) && strncmp(gh.cVersion,"87a",3))) {
		MessageBox(hMainDlg,"Not a GIF file, or incorrect version number.",0,MB_ICONEXCLAMATION);
		CloseHandle(hFile);
		return;
	}
	ReadFile(hFile,&glsd,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR),&dw,0);
	if (glsd.iGctFlag) SetFilePointer(hFile,iGctSize[glsd.iSizeOfGct],0,FILE_CURRENT);
	for (;;) {
		ReadFile(hFile,&c,1,&dw,0);
		if (dw==0) {
			MessageBox(hMainDlg,"Premature end of file encountered; no GIF image data present.",0,MB_ICONEXCLAMATION);
			CloseHandle(hFile);
			return;
		}
		if (c==0x2C) {
			ReadFile(hFile,&gid,sizeof(GIF_IMAGE_DESCRIPTOR),&dw,0);
			break;
		} else if (c==0x21) {
			ReadFile(hFile,&c,1,&dw,0);
			if (c==0xF9) {
				SetFilePointer(hFile,sizeof(GIF_GRAPHIC_CONTROL_EXTENSION),0,FILE_CURRENT);
			} else {
				for (;;) {
					ReadFile(hFile,&c,1,&dw,0);
					if (!c) break;
					SetFilePointer(hFile,c,0,FILE_CURRENT);
				}
			}
		}
	}
	CloseHandle(hFile);
	sprintf(sz,"%d x %d\n%s (%d bpp)",glsd.iWidth,glsd.iHeight,gid.iLctFlag?"local":(glsd.iGctFlag?"global":"none"),gid.iLctFlag?(gid.iSizeOfLct+1):(glsd.iSizeOfGct+1));
	SetDlgItemText(hMainDlg,IDC_STATIC_MAIN_INFO,sz);
}

void MakeGIF()
{
	OPENFILENAME ofn;
	char szFileName[MAX_PATH];
	DWORD dwIndex, dw;
	char szColorTable[768];
	unsigned char c;
	HANDLE hFileOut, hFileIn;
	GIF_HEADER gh;
	GIF_LOGICAL_SCREEN_DESCRIPTOR glsd1, glsd;
	GIF_GRAPHIC_CONTROL_EXTENSION ggce;
	GIF_IMAGE_DESCRIPTOR gid;
	ZeroMemory(&glsd1,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR));
	szFileName[0]=0;
	ofn.lStructSize=sizeof(ofn);
	ofn.hwndOwner=hMainDlg;
	ofn.hInstance=hInst;
	ofn.lpstrFilter="GIF Files (*.gif)\0*.gif\0";
	ofn.lpstrCustomFilter=NULL;
	ofn.nMaxCustFilter=0;
	ofn.nFilterIndex=1;
	ofn.lpstrFile=szFileName;
	ofn.nMaxFile=MAX_PATH;
	ofn.lpstrFileTitle=NULL;
	ofn.nMaxFileTitle=0;
	ofn.lpstrInitialDir=NULL;
	ofn.lpstrTitle="Make Animated GIF";
	ofn.Flags=OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_PATHMUSTEXIST;
	ofn.nFileOffset=NULL;
	ofn.nFileExtension=NULL;
	ofn.lpstrDefExt="gif";
	ofn.lCustData=NULL;
	ofn.lpfnHook=NULL;
	ofn.lpTemplateName=NULL;
	if (GetSaveFileName(&ofn)) {
		hFileOut=CreateFile(szFileName,GENERIC_WRITE,FILE_SHARE_READ,0,CREATE_ALWAYS,0,0);
		strncpy((char *)&gh,"GIF89a",6);
		WriteFile(hFileOut,&gh,sizeof(GIF_HEADER),&dw,0);
		WriteFile(hFileOut,&glsd1,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR),&dw,0);
		if (IsDlgButtonChecked(hMainDlg,IDC_CHECK_MAIN_LOOP)) {
			WriteFile(hFileOut,"\41\377\013NETSCAPE2.0\003\001\377\377\0",19,&dw,0);
		}
		for (dwIndex=0;dwIndex<(unsigned)SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_GETCOUNT,0,0);dwIndex++) {
			ZeroMemory(&glsd,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR));
			ZeroMemory(&ggce,sizeof(GIF_GRAPHIC_CONTROL_EXTENSION));
			ZeroMemory(&gid,sizeof(GIF_IMAGE_DESCRIPTOR));
			hFileIn=CreateFile((char *)SendDlgItemMessage(hMainDlg,IDC_LIST_MAIN_FRAMES,LB_GETITEMDATA,dwIndex,0),GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_SEQUENTIAL_SCAN,0);
			if (hFileIn==INVALID_HANDLE_VALUE) {
				MessageBox(hMainDlg,"Could not open file. GIF creation aborted.",0,0);
				CloseHandle(hFileOut);
				return;
			}
			ReadFile(hFileIn,&gh,sizeof(GIF_HEADER),&dw,0);
			if (strncmp(gh.cSignature,"GIF",3) || (strncmp(gh.cVersion,"89a",3) && strncmp(gh.cVersion,"87a",3))) {
				MessageBox(hMainDlg,"Not a GIF file, or incorrect version number.",0,MB_ICONEXCLAMATION);
				CloseHandle(hFileIn);
				CloseHandle(hFileOut);
				return;
			}
			ReadFile(hFileIn,&glsd,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR),&dw,0);
			if (glsd.iGctFlag) ReadFile(hFileIn,szColorTable,iGctSize[glsd.iSizeOfGct],&dw,0);
			if (glsd1.iWidth<glsd.iWidth) glsd1.iWidth=glsd.iWidth;
			if (glsd1.iHeight<glsd.iHeight) glsd1.iHeight=glsd.iHeight;
			for (;;) {
				ReadFile(hFileIn,&c,1,&dw,0);
				if (dw==0) {
					MessageBox(hMainDlg,"Premature end of file encountered; no GIF image data present.",0,MB_ICONEXCLAMATION);
					CloseHandle(hFileIn);
					CloseHandle(hFileOut);
					return;
				}
				if (c==0x2C) {
					ReadFile(hFileIn,&gid,sizeof(GIF_IMAGE_DESCRIPTOR),&dw,0);
					if (gid.iLctFlag) {
						ReadFile(hFileIn,szColorTable,iGctSize[gid.iSizeOfLct],&dw,0);
					} else {
						gid.iLctFlag=1;
						gid.iSizeOfLct=glsd.iSizeOfGct;
					}
					break;
				} else if (c==0x21) {
					ReadFile(hFileIn,&c,1,&dw,0);
					if (c==0xF9) {
						ReadFile(hFileIn,&ggce,sizeof(GIF_GRAPHIC_CONTROL_EXTENSION),&dw,0);
					} else {
						for (;;) {
							ReadFile(hFileIn,&c,1,&dw,0);
							if (!c) break;
							SetFilePointer(hFileIn,c,0,FILE_CURRENT);
						}
					}
				}
			}
			ggce.iBlockSize=4;
			ggce.iDelayTime=(unsigned short)GetDlgItemInt(hMainDlg,IDC_EDIT_MAIN_DELAY,0,0);
			ggce.iDisposalMethod=2;
			c=(char)0x21;
			WriteFile(hFileOut,&c,1,&dw,0);
			c=(char)0xF9;
			WriteFile(hFileOut,&c,1,&dw,0);
			WriteFile(hFileOut,&ggce,sizeof(GIF_GRAPHIC_CONTROL_EXTENSION),&dw,0);
			c=(char)0x2C;
			WriteFile(hFileOut,&c,1,&dw,0);
			WriteFile(hFileOut,&gid,sizeof(GIF_IMAGE_DESCRIPTOR),&dw,0);
			WriteFile(hFileOut,szColorTable,iGctSize[gid.iSizeOfLct],&dw,0);
			ReadFile(hFileIn,&c,1,&dw,0);
			WriteFile(hFileOut,&c,1,&dw,0);
			for (;;) {
				ReadFile(hFileIn,&c,1,&dw,0);
				WriteFile(hFileOut,&c,1,&dw,0);
				if (!c) break;
				ReadFile(hFileIn,szColorTable,c,&dw,0);
				WriteFile(hFileOut,szColorTable,c,&dw,0);
			}
			CloseHandle(hFileIn);
		}
		c=(char)0x3B;
		WriteFile(hFileOut,&c,1,&dw,0);
		SetFilePointer(hFileOut,6,0,FILE_BEGIN);
		WriteFile(hFileOut,&glsd1,sizeof(GIF_LOGICAL_SCREEN_DESCRIPTOR),&dw,0);
		CloseHandle(hFileOut);
	}
}